% This is a preliminary implementation
% of the method described in the paper
% "Implicit QR with compression",
% by P. Boito, Y. Eidelman, L. Gemignani
% and I. Gohberg.
%
% The proposed method is a variant of
% structured implicit QR, where the
% unitary factors V and F are not explicitly
% used; instead, the generators are
% updated directly and then compressed.
%
% In this version we implement double shift.
%
% Input data: pol (vector of polynomial coefficients).
% Output data: ev (vector of computed eigenvalues/roots),
%              ev_err (absolute forward error in max norm),
%              av_it (average number of iterations per eigenvalue).
%
% Last update by Paola, January 30, 2012.
%

function [ev,ev_err,av_it]=qrwc_ds(pol)

% write pol in column form if necessary
[pol1,pol2]=size(pol);
if pol1<pol2
    pol=pol.';
end

n=length(pol)-1;
ev=zeros(n,1); % computed eigenvalues will be stored here
iterations=zeros(n,1);  % number of iterations needed for each eigenvalue
%delta_vec=ones(n-1,1);

dat1=0.75;
dat2=-0.4375;  % to be used in exceptional shift

count_oldQR=0; % count number of calls to old QR (without compression)
count_es=0;  % count number of exceptional shifts

%*************************************************
% Initialization of F, V, p, q.
%
% F is 3x3x(n-1),
% V is 2x2x(n-1),
% p and q are vectors of length n
%*************************************************

E2=eye(2);
P3=[0 1 0; 1 0 0; 0 0 1];

F=zeros(3,3,n-1);

F(:,:,1:n-3)=P3(:,:,ones(1,n-3)); % initialize F with permuted identity matrices

F(:,:,n-2)=[0 0 1; 1 0 0; 0 1 0];
F(:,:,n-1)=eye(3);

V=E2(:,:,ones(1,n-1));  % initialize V with identity matrices

q=pol(2:n+1);   % initialize p and q with polynomial coefficients
q(n)=q(n)+1;
p=[1;zeros(n-1,1)];

%*************************************************
% Initialization of g, h , b, sigma.
%
% g is nx2,
% h is 2xn,
% b is 2x2xn,
% sigma is a vector of length n-1.
%*************************************************

% Note that generators of index n have smaller size
% (extra entries are set to zero)
h=zeros(2,n);
g=zeros(n,2);
b=zeros(2,2,n-1);
sigma=zeros(n-1,1);
gamma=zeros(n-1,2);

h(:,1:n-2)=F(1:2,1,1:n-2);
b(:,:,1:n-2)=F(1:2,2:3,1:n-2);

h(:,n-1)=F(1:2,1,n-1);
b(:,:,n-1)=[F(1:2,2,n-1), zeros(2,1)];
h(:,n)=[1; 0];

gamma(1,:)=[0 1];
g(1,:)=[1 0];

for k=1:n-2
    M=V(:,:,k+1)*[gamma(k,:), 0; 0, 0, 1]*F(:,:,k);
    sigma(k)=M(1,1);
    g(k+1,:)=M(1,2:3);
    gamma(k+1,:)=M(2,2:3);
end
sigma(n-1)=gamma(n-1,:)*h(:,n-1);
g(n,:)=gamma(n-1,:)*b(:,:,n-1);


% *** CHECK: compute the companion matrix and verify that generators are
% correct.
 CM=compan(pol);
 CMgen=compute_matrix(g,h,b,sigma,p,q);
 norm(CM-CMgen)
% end check

defl_size=n; % this means we are working on a matrix of size defl_size

while defl_size>2   % begin main loop

    % Compute diagonal and subdiagonal entries
    % (subdiag2 is the vector of subdiagonal entries computed
    % directly from delta)
    diagonals=g(:,1).*h(1,:).'+g(:,2).*h(2,:).'-p.*q;
    %subdiag2=ones(defl_size-1,1);
    %subdiag2(1:defl_size-2)=delta_vec(1:defl_size-2);
    %if defl_size==n
    % subdiag2(defl_size-1)=sigma(defl_size-1)-p(defl_size).*q(defl_size-1);
    %else
    %    subdiag2(defl_size-1)=delta_vec(defl_size-1);
    %end
    subdiag=sigma(1:defl_size-1)-p(2:defl_size).*q(1:defl_size-1);

    %display(subdiag)
    % display([subdiag,subdiag2])
    %subdiag=subdiag2;

    % Look for first (from the bottom) small subdiagonal entry
    % (standard deflation criterion)
    m=defl_size-1;
    while (m>0)&&(abs(subdiag(m))>5*eps*(abs(diagonals(m))+abs(diagonals(m+1))))
        m=m-1;
    end
    % we'll do this somewhat Fortran-style sometimes...
    L=zeros(2);
    L(1,1)=diagonals(defl_size-1);
    L(2,2)=diagonals(defl_size);
    L(2,1)=subdiag(defl_size-1);
    L(1,2)=g(defl_size-1,:)*b(:,:,defl_size-1)*h(:,defl_size)-p(defl_size-1)*q(defl_size);
    s=abs(L(1,1))+abs(L(1,2))+abs(L(2,1))+abs(L(2,2));
    % Has an eigenvalue been found?
    if m==defl_size-1
        fprintf('one eigenvalue found \n')
        ev(defl_size)=diagonals(defl_size);
        defl_size=defl_size-1;
        display(defl_size)
        % Has a pair of complex conjugate eigenvalues been found?
    else
        if m==defl_size-2
            fprintf('two eigenvalues found \n')
            ev(defl_size-1:defl_size)=eig(L);
            defl_size=defl_size-2;
            display(defl_size)
        else
            % we'll have to perform QR then!
            iterations(defl_size)=iterations(defl_size)+1;
            if iterations(defl_size)>10 % exceptional shift if needed
                L(1,1)=dat1*s+L(2,2);
                L(1,2)=dat2*s;
                L(2,1)=s;
                L(2,2)=L(1,1);
                iterations(defl_size)=0;
                count_es=count_es+1;
                fprintf('Exceptional shift...\n')
            end
            m=m+1;

            a_final=L;

            % Check for "magic trick" (two small consecutive subdiagonal entries)
            % in the submatrix of indices m:defl_size
            %

           

            k=defl_size-2; % now assume that the bulge-chasing process starts from k...
            while k>m

                nrm=zeros(3,1);
               % a_final=[diagonals(defl_size-1), g(defl_size-1,:)*b(:,:,defl_size-1)*h(:,defl_size)-p(defl_size-1)*q(defl_size);...
               %     subdiag(defl_size-1), diagonals(defl_size)];

                a_initial=[diagonals(k), g(k,:)*b(:,:,k)*h(:,k+1)-p(k)*q(k+1); subdiag(k), diagonals(k+1)];
                nrm(1)=((a_final(2,2)-a_initial(1,1))*(a_final(1,1)-a_initial(1,1))-a_final(1,2)*a_final(2,1))/a_initial(2,1)+a_initial(1,2);
                nrm(2)=a_initial(2,2)-a_initial(1,1)-(a_final(2,2)-a_initial(1,1))-(a_final(1,1)-a_initial(1,1));
                nrm(3)=subdiag(k+1);
                % normalize nrm as suggested in Numerical Recipes
                nrs=abs(nrm(1))+abs(nrm(2))+abs(nrm(3));
                v_vec=nrm/nrs;
                % Let's try the NR criterion:
                nrc=eps*abs(nrm(1))*(abs(diag(k+1))+abs(diag(k))+abs(diag(k-1)));
                if abs(subdiag(k-1))*(abs(nrm(2))+abs(nrm(3)))<nrc
                    fprintf('Magic trick!\n')
                    break
                end

                % Other criterion for magic trick:
                %G=hh3(nrm);
                %            G=hh3(v_vec);
                %            y=[subdiag(k-1);0;0];
                %            y=G*y;

                %            if norm(y(2:3))<threshold
                %               fprintf('Magic trick!\n')
                %              break
                %            end

                % And yet another one...

                %            term1=abs(subdiag(k-1))*(abs(v_vec(2))+abs(v_vec(3)));
                %            term2=eps*abs(v_vec(1))*(abs(diagonals(k-1))+abs(diagonals(k))+abs(diagonals(k+1)));

                %            if term1<=term2
                %                fprintf('Magic trick!\n')
                %                break
                %            end
                k=k-1;
            end
            m=k;
            nrm=zeros(3,1);
            % a_final=[diagonals(defl_size-1), g(defl_size-1,:)*b(:,:,defl_size-1)*h(:,defl_size)-p(defl_size-1)*q(defl_size);...
            %    subdiag(defl_size-1), diagonals(defl_size)];
            a_initial=[diagonals(m), g(m,:)*b(:,:,m)*h(:,m+1)-p(m)*q(m+1); subdiag(m), diagonals(m+1)];
            nrm(1)=((a_final(2,2)-a_initial(1,1))*(a_final(1,1)-a_initial(1,1))-a_final(1,2)*a_final(2,1))/a_initial(2,1)+a_initial(1,2);
            nrm(2)=a_initial(2,2)-a_initial(1,1)-(a_final(2,2)-a_initial(1,1))-(a_final(1,1)-a_initial(1,1));
            nrm(3)=subdiag(m+1);
            % normalize nrm as suggested in NR
            nrs=abs(nrm(1))+abs(nrm(2))+abs(nrm(3));
            v_vec=nrm/nrs;

            % We are ready now to perform QR on the submatrix of indices
            % m:defl_size.
            %
            % **********************************************************
            % Choose criterion for switching to QR without compression:
            % uncomment the following to use the criterion based on the 
            % product of quasiconsecutive subdiagonal entries:
            
            %if (defl_size>4)&&(abs(subdiag(defl_size-1)*subdiag(defl_size-3))>10*eps)&&(abs(subdiag(defl_size-2)*subdiag(defl_size-4))>10*eps)
                [g,h,b,sigma,p,q,delta_vec,CM]=bulge_chasing_routine(g,h,b,sigma,p,q,m,defl_size,v_vec,n,CM,pol);
            %else
            %    if (defl_size<=4)
            %        [g,h,b,sigma,p,q,V,F,delta_vec,CM]=bulge_chasing_VF(g,h,b,sigma,p,q,m,defl_size,v_vec,n,CM,pol);
            %    else
            %        %[g,h,b,sigma,p,q,V,F,CM]=...
            %        %    bc_ds_alancienne(g,h,b,sigma,p,q,V,F,n,defl_size,m,v_vec,CM);
            %        [g,h,b,sigma,p,q,V,F,delta_vec,CM]=bulge_chasing_VF_update(g,h,b,sigma,p,q,m,defl_size,v_vec,n,CM,pol);
            %        count_oldQR=count_oldQR+1;
            %    end
            %end
            
            %... or uncomment the following to use criterion based on
            % block structure of F (i.e. if F=[0 x x; -1 0 0; 0 x x];
            % note that F(1,1)=0 always in this implementation):
            
            %F_threshold=1000*eps;
            %if (abs(F(3,1,defl_size-1))<F_threshold)||(abs(F(3,1,defl_size-2))<F_threshold)
            %    [g,h,b,sigma,p,q,V,F,CM]=...
            %            bc_ds_alancienne(g,h,b,sigma,p,q,V,F,n,defl_size,m,v_vec,CM);
            %else
            %     [g,h,b,sigma,p,q,V,F,delta_vec,CM]=bulge_chasing_VF(g,h,b,sigma,p,q,m,defl_size,v_vec,n,CM,pol);
            %end

            % end of criterion choice
            % ***********************************************************
                
            
        end
    end
end
if defl_size==2
    L=zeros(2);
    L(1,1)=diagonals(defl_size-1);
    L(2,2)=diagonals(defl_size);
    L(2,1)=subdiag(defl_size-1);
    L(1,2)=g(defl_size-1,:)*b(:,:,defl_size-1)*h(:,defl_size)-p(defl_size-1)*q(defl_size);
    ev(1:2)=eig(L);
else
    ev(1)=diagonals(1);
end
ev_err=AbsDistance(abs(ev), abs(roots(pol)));
av_it=(10*count_es+sum(iterations))/n;
fprintf('Error on computed eigenvalues: %g\n', ev_err);
fprintf('Average number of iterations: %f\n',av_it);
%fprintf('Number of calls to QR without compression: %d\n',count_oldQR)